/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.healing;

import java.util.ArrayList;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.AbstractCheck;
import de.hft.stuttgart.citydoctor2.check.HealingID;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.error.AttributeMissingError;
import de.hft.stuttgart.citydoctor2.datastructure.AbstractBuilding;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.Installation;
import de.hft.stuttgart.citydoctor2.datastructure.BuildingPart;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.GmlElement;
import de.hft.stuttgart.citydoctor2.datastructure.LinkedPolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;

public class HealMissingSolid implements HealingMethod {
	
	private static final Logger logger = LogManager.getLogger(HealMissingSolid.class);

	@Override
	public HealingMethod createNew() {
		return new HealMissingSolid();
	}

	@Override
	public boolean visit(AttributeMissingError err, ModificationListener l) {
		logger.debug("Executing Repair for AttributeMissingError");
		if (err.getNameOfAttribute().equalsIgnoreCase("lod1solid")) {
			return createSolid(err.getFeature(), Lod.LOD1);
		} else if (err.getNameOfAttribute().equalsIgnoreCase("lod2solid")) {
			return createSolid(err.getFeature(), Lod.LOD2);
		} else if (err.getNameOfAttribute().equalsIgnoreCase("lod3solid")) {
			return createSolid(err.getFeature(), Lod.LOD3);
		} else if (err.getNameOfAttribute().equalsIgnoreCase("lod4solid")) {
			return createSolid(err.getFeature(), Lod.LOD4);
		} else if (err.getNameOfAttribute().equalsIgnoreCase("any solid")) {
			return createSolid(err.getFeature());
		}
		return false;
	}

	private boolean createSolid(GmlElement feature) {
		if (feature instanceof AbstractBuilding ab) {
			List<Polygon> findPolygons = collectPolygons(ab);
			if (findPolygons.isEmpty()) {
				return false;
			}
			Lod highestLod = Lod.LOD1;
			for (Polygon p : findPolygons) {
				if (p.getParent().getLod().isHigher(highestLod)) {
					highestLod = p.getParent().getLod();
				}
			}
			return createSolid(feature, highestLod);
		}
		return false;
	}

	private boolean createSolid(GmlElement feature, Lod lod) {
		if (feature instanceof Building building) {
			boolean createdSolidInMainFeature = createSolidInFeature(lod, building);
			boolean hasCreatedSolid = createdSolidInMainFeature;
			for (BuildingPart part : building.getBuildingParts()) {
				boolean createdSolidInPart = createSolidInFeature(lod, part);
				hasCreatedSolid = hasCreatedSolid || createdSolidInPart;
			}
			return hasCreatedSolid;
		} else if (feature instanceof BuildingPart part) {
			return createSolidInFeature(lod, part);
		}
		return false;
	}
	
	private List<Polygon> collectPolygons(CityObject co) {
		List<Polygon> polygons = new ArrayList<>();
		AbstractCheck polygonCheck = new AbstractCheck() {
			@Override
			public void check(Polygon p) {
				if (!p.isLink()) {
					polygons.add(p);
				}
			}
		};
		co.accept(polygonCheck);
		return polygons;
	}

	private boolean createSolidInFeature(Lod lod, AbstractBuilding building) {
		Geometry geom = new Geometry(GeometryType.SOLID, lod);
		List<Polygon> polygons = collectPolygons(building, lod, geom);
		for (Polygon p : polygons) {
			geom.addPolygon(p);
		}
		if (geom.getPolygons().isEmpty()) {
			// no geometry has been created
			return false;
		}
		// remove both solid and multisurface geometry
		building.removeGeometry(lod);
		building.addGeometry(geom);
		return true;
	}

	private List<Polygon> collectPolygons(AbstractBuilding feature, Lod lod, Geometry solidGeometry) {
		List<Polygon> polygons = new ArrayList<>();
		// collect all concrete polygons from main geometry if any
		for (Geometry geom : feature.getGeometries()) {
			if (geom.getLod().equals(lod)) {
				for (Polygon p : geom.getPolygons()) {
					if (!p.isLink()) {
						// concrete polygon, add to list
						polygons.add(p);
					}
				}
			}
		}

		for (BoundarySurface bs : feature.getBoundarySurfaces()) {
			collectPolygons(bs, lod, polygons, solidGeometry);
		}
		for (Installation bi : feature.getBuildingInstallations()) {
			collectPolygons(bi, lod, polygons, solidGeometry);
			for (BoundarySurface bs : bi.getBoundarySurfaces()) {
				collectPolygons(bs, lod, polygons, solidGeometry);
			}
		}
		return polygons;
	}

	/**
	 * Only find concretePolygons and add them as link to the list
	 * 
	 * @param co
	 * @param lod
	 * @param polygons
	 */
	private void collectPolygons(CityObject co, Lod lod, List<Polygon> polygons, Geometry solidGeometry) {
		for (Geometry geom : co.getGeometries()) {
			if (geom.getLod().equals(lod)) {
				for (Polygon p : geom.getPolygons()) {
					if (p instanceof ConcretePolygon cp) {
						// concrete polygon, create link and add to list
						polygons.add(new LinkedPolygon(cp, solidGeometry));
					}
				}
			}
		}
	}

	@Override
	public HealingID getID() {
		return HealingID.SE_MISSING_LOD2_SOLID;
	}

}
